#!/usr/bin/perl
#
# Combine missing glyphs in 7x14.bdf from 7x13.bdf
#
# Usage:
# $0 -a ja ./7x14.bdf ./7x13.bdf > 7x14ja.bdf
# $0 -a ja ./7x14B.bdf -b ./7x14ja.bdf > 7x14jaB.bdf
#

$bolden = 0;
$add_style = "";

$usage = "Usage: perl this.pl ./7x14.bdf ./7x13.bdf > 7x14new.bdf";
$charsdir = "./chars$$";	#tmpdir

# first, open 7x14.bdf and divide glyphs out into ./chars/NNNNN
if ($ARGV[0] eq "-a") {shift; $add_style = shift;}
if ($ARGV[0] eq "-b") {$bolden = 1; shift;}
$bdf = shift(@ARGV) || die $usage;
open(BDF, "<", $bdf);
$state = "header";
-d $charsdir || mkdir($charsdir) || die "cannot create ./chars/";
open(META, ">", "$charsdir/+HEADER") || die "cannot write +HEADER";
while(<BDF>) {
	if (/^STARTCHAR/) {
		undef $code;
		$cspec = $_;
		# read until ENDCHAR
		while(<BDF>) {
			if (/^ENCODING ([0-9]+)/) { $code = $1; }
			$cspec .= $_;
			last if /^ENDCHAR/;
		}
		if (! defined($code)) { die "undefined ENCODING in STARTCHAR"; }
		# dump to a separate file
		open(CSPEC,">", sprintf("$charsdir/%05s",$code));
		print CSPEC $cspec;
		close(CSPEC);

		# state transition
		if ($state eq "header") {
			close(META);
			open(META, ">", "$charsdir/=TRAILER") || die "cannot write =TRAILER";
			$state = "trailer";
		}

		next;
	}

	print META $_;
}
close BDF;
close META;

# next, open 7x13.bdf and dump it, but
# - don't overwrite existing ./char/NNNNN
# - add a scanline in BITMAP to make it 7x14
# - don't write out headers
if ($ARGV[0] eq "-b") {$bolden = 1; shift;}
$bdf = shift(@ARGV) || die $usage;
open(BDF, "<", $bdf);
$state = "header";
#open(META, ">", "$charsdir/+HEADER") || die "cannot write +HEADER";
while(<BDF>) {
	if (/^STARTCHAR/) {
		undef $code;
		$cspec = $_;
		$in_bitmap=0;
		$add_scanline=0;
		# read until ENDCHAR
		while(<BDF>) {
			# edit
			if (/^ENCODING ([0-9]+)/) { $code = $1;}
			s/^SWIDTH [0-9]+ /SWIDTH 960 /;
			if (/^BBX ([0-9]+)/) { $bbw=$1; }
			if (s/^BBX 7 13 /BBX 7 14 /) { $add_scanline=1; }
			s/^BITMAP/$&."\n00"/e if $add_scanline;
			$in_bitmap=0 if /^ENDCHAR/;
			if ($bolden && $in_bitmap && /^[0-9A-F]+/) {
				$digits=length($&);
				$right_pad_bits = length($&)*4 - $bbw;
				$o = hex("0x" . $&);
				# overstrike to right side
				$o = ($o/2) | $o;
				# limit to BBw bits; MSB packed!
				$o &= (((1<<$bbw)-1) << $right_pad_bits);
				$_ = sprintf("%0*X\n", $digits, $o);
			}
			$cspec .= $_;
			$in_bitmap = 1 if /^BITMAP/;
			last if /^ENDCHAR/;
		}
		if (! defined($code)) { die "undefined ENCODING in STARTCHAR"; }

		# dump to a separate file
		# don't overwrite
		if ( ! -e sprintf("$charsdir/%05s",$code)) {
		open(CSPEC,">", sprintf("$charsdir/%05s",$code));
		print CSPEC $cspec;
		close(CSPEC);
		}

	}

	#print META $_;
}
close BDF;

# re-count the chars
$chars = 0;
opendir CHARS, "$charsdir" || die "no ./chars?";
while ($_ = readdir CHARS) { $chars++ if /^[0-9]+/; }
#$chars = scalar(grep {/^[0-9]+/} readdir CHARS); #slower list context
closedir CHARS;

# re-read the header, and modify
open(META,"<","$charsdir/+HEADER");
{
	local $/;	# full slurp mode
	$_ = <META>;
}
close META;
s/^FONT (-Misc-Fixed)-(\w+)-(R-Normal-)(-14-130-75-75-C-70-ISO10646-1)/
"COMMENT \"This font is 7x14 with missing glyphs filled from 7x13\"\n" .
"FONT $1-".(($bolden)?"Bold":"${2}")."-${3}${add_style}${4}"/eim;
s/^ADD_STYLE_NAME .*/"ADD_STYLE_NAME \"$add_style\""/em;
s/^CHARS [0-9]+/"CHARS $chars"/em;
if (/^WEIGHT_NAME .*/m && $bolden) { s//WEIGHT_NAME "Bold"/m; }
open(META,">","$charsdir/+HEADER");
print META $_;
close META;

# output the result in +HEADER, NNNNN, =TRAILER order
opendir CHARS, "$charsdir" || die "no ./chars?";
foreach my $f (sort grep {!/^\./} readdir CHARS) {
	local $/;
	open F,"<","$charsdir/$f";
	print <F>;
	close F;
}
closedir CHARS;

# cleanup
system("rm -fr $charsdir");
